#ifndef BF_DLFCN_H
#define BF_DLFCN_H

/* PORTABILITY
 *
 * Modern UNIX/Linux systems should have a dlfcn.h header file, as specified
 * by POSIX POSIX 1003.1-2003 and SUSv2, to provide an interface to the
 * dynamic linker.  If your system does not have it, then we will use this
 * file, which provides macros that always indicate failures.
 *
 * We use the dynamic loader functionality for several things:
 * - bfcryptloader.c uses it on UNIX/Linux to load a custom encryption module.
 * - bfpwdlocloader.c uses it on UNIX/Linux to load a custom password locator
 *			module.
 * - platform/sysconf_mem.c uses it on AIX if the vmgetinfo function was not
 *			found by configure, because it might still be available in the
 *			kernel.  This is needed to get the MEM_* built-in collector
 *			values, because AIX doesn't support the _SC_PHYS_PAGES and
 *			_SC_AVPHYS_PAGES sysconf queries.
 * - platform.c uses it on UNIX/Linux to support the undocumented feature of
 *			creating your own dot commands, which has not to my knowledge
 *			ever been used.  Look at bf_dynamic_function if you are curious.
 *
 * It goes without saying that if we have to use this file during compilation
 * because we can't find a usable dlfcn.h file on your system, then none of
 * the features listed above will work.
 */

#undef RTLD_LAZY
#undef RTLD_NOW
#undef RTLD_GLOBAL
#undef RTLD_LOCAL
#undef dlopen
#undef dlclose
#undef dlsym
#undef dlerror

#define RTLD_LAZY (0)
#define RTLD_NOW (0)
#define RTLD_GLOBAL (0)
#define RTLD_LOCAL (0)
#define dlopen(path,mode) (NULL)
#define dlclose(handle) (-1)
#define dlsym(handle,sym) (NULL)
#define dlerror() ("Dynamic loader functions do not appear to be supported on this operating system.")

#endif /* BF_DLFCN_H */

