/*
Copyright (C)2008-2011, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#ifndef BFCONF_H
#define BFCONF_H

/*
 * In the 3.x line, misc.c has a function called getConfig that opens the
 * bfagent.conf file and reads it in, looking for a specific item, and then
 * closes it again.  No attempt was made to cache this information, so the
 * file would be read and reread countless times unnecessarily.  This file
 * provides functions for caching this data in a hash table where it can be
 * accessed quickly and refreshed only when needed.
 */

void bfconf_refresh(void);
void bfconf_set_path(const char *path);
const char *bfconf_get_path(void);
int bfconf_get_errno(void);

int bfconf_dump(void);
char *bfconf_get(const char *key, const char *def);
int bfconf_exists(const char *key);

typedef struct {
	int activity_log:1;
	int cc_suppress_server_root:1;
	int cygwin:1;
	int disable_cleanup_during_exit:1;
	int disable_telnet_support:1;
	int disable_transcode:1;
	int enable_agent_dll:1;
	int enable_credential_retention:1;
	int enable_deprecated_path_mapping:1;
	int fail_if_dot_source_fails:1;
	int fail_if_log_file_fails:1;
	int getaddrinfo_use_addrconfig:1;
	int ignore_clearcase_view:1;
	int ipv6_connect_first_addr_only:1;
	int leave_tmp_file:1;
	int map_drive_is_failure:1;
	int nologonshell:1;
	int no_lang_env:1;
	int no_preparse_command:1;
	int no_pty:1;
	int shellarg:1;
	int shell_compatible_undef_vars:1;
	int win_reexec_after_auth:1;
	int password_encrypt_module:1;
	int password_encrypt_config:1;
	int xstream_allow_ssl_mismatch:1;
	int xstream_debug:1;
	int xstream_disable_md5:1;
	int xstream_listen_randomize:1;
} bfconf_data_t;

#ifndef BFCONF_C
extern bfconf_data_t bfconf_data;
#endif

#define BFCONF_EXISTS(x) (bfconf_data.x)

#endif /* BFCONF_H */

