/*
Copyright (C)2008-2011, IBM, Inc. All rights reserved.
IBM, Inc. owns the copyright and other intellectual
property rights in this software. Duplication or use of the
Software is not permitted except as expressly provided in a
written agreement between your company and IBM, Inc.
*/

/* Clearly it is not acceptable for the agent to prompt for passwords in
 * a functional system.  This file exists purely as an example to
 * illustrate the calling convention that a custom password locator would
 * need to honor.
 */

#include "bf_agent.h"

static const char *MODULE_NAME = "bfpwdpromptloc";

static const char *getName(void) { return MODULE_NAME; }

static char *getPassword(const char *prop, const char *location) {
	char *password = malloc(512);
	char *p;

	if (NULL == password) {
		fprintf(stdout, "Error: out of memory allocating password buffer\n");
		fflush(stdout);
		return NULL;
	}

	fprintf(stdout, "Enter password for property \"%s->%s\": ", prop, location);
	fflush(stdout);
	if (NULL == fgets(password, 512, stdin)) {
		fprintf(stdout, "\nError: EOF or error condition on stdin.\n");
		fflush(stdout);
		free(password);
		return NULL;
	}

	/* Ensure proper termination and strip any trailing control chars */
	password[511] = '\0';
	p = &password[(int)strlen(password)];
	while (--p >= password) {
		if ((unsigned char)(*p) >= 0x20) { break; }
		*p = '\0';
	}
	return password;
}

static const passwd_locator_ops my_ops = {
	&getName,
	&getPassword
};


#ifdef WIN32
__declspec(dllexport)
#endif
const passwd_locator_ops *bfpwdpromptloc_init(void) { return &my_ops; }

