/*
Copyright (C)2007-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#ifndef BUFFER_H
#define BUFFER_H

/* DBCS-safe snprintf-like functions and self-expanding
 * string buffers.
 */

#include "bf_agent.h"

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

int format_message(char *, const char *, ...);

/* Akin to snprintf */
char *buffer_format(
		size_t *lenptr,
		const char *bucket,
		const char *msgkey,
		const char *fmt,
		...);

/* Akin to vsnprintf */
char *buffer_format_v(
		size_t *lenptr,
		const char *bucket,
		const char *msgkey,
		const char *fmt,
		va_list va);

/* Expands the buffer, if possible.  Note: caller must the return value
 * for safety; failure is indicated by a ret code of 0.  It is preferable
 * to use the BUFFER_ENSURE_CAPACITY macro for convenience.
 */
int buffer_ensure_capacity(char **bufptr, size_t *lenptr, size_t size);

/* This macro calls buffer_ensure_capacity and does a "goto out_of_memory;"
 * if the return value indicates failure.  Since this will fail to compile
 * unless you provide that goto label for error handling, it is the
 * preferred way to check a buffer's capacity.
 */
#define BUFFER_ENSURE_CAPACITY(bufptr, lenptr, size) do { \
	if (!buffer_ensure_capacity(bufptr, lenptr, size)) { \
		goto out_of_memory; \
	} \
} while (0)


/* ZOS is a bit more complicated due to its EBCDIC-ness.  Don't
 * call either of these directly; use buffer_append_escaped.
 */
#ifdef ZOS
int buffer_append_escaped_zos(
		char **bufptr,
		size_t *blenptr,
		size_t *bidxptr,
		const char *s);
#define buffer_append_escaped buffer_append_escaped_zos
#else /* !ZOS */
int buffer_append_escaped_utf8(
		char **bufptr,
		size_t *blenptr,
		size_t *bidxptr,
		const char *s);
#define buffer_append_escaped buffer_append_escaped_utf8
#endif /* ZOS/! */

#endif /* BUFFER_H */

