/*
Copyright (C)2003-2014, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/* Implementation for "cmd exec" and "cmd expand", the primary means of
 * executing commands through the agent.
 */

extern bf_internal_t internals[];
extern char *sourcelist;

#define DEFAULT_TIMEOUT 300

int ParseNum(const char c)
{
	int ret = 0;

	if (c > 57) {
		ret = c - 87;
	} else {
		ret = c - 48;
	}

	return ret;
}

char *Parse(const char *str)
{
	char *result, *p, *q;
	char s[9], t[2];
	int len, i, j, n;

	len = strlen(str);
	result = (char *)calloc(len, sizeof(char));
	p = (char *)strstr(str, FORGE);
	len = p - str + 1;
	q = (char *)calloc(len, sizeof(char));
	strncpy(q, str, (len - 1));
	p += 5;
	strncpy(s, p, (sizeof(s) - 1));
	s[sizeof(s)-1] = '\0';

#ifdef ZOS
	convert_etoa(s);
#endif /* ZOS */

	len = strlen(q);
	t[1] = 0;
	for (i = 0, j = 0; i < len; i++, j++) {
		t[0] = *q;
		if ('^' == t[0]) {
			t[0] = *(++q);
			n = atoi(t);
			i += 1 + n;
			t[0] = *(++q);
#ifdef ZOS
			convert_etoa(t);
#endif /* ZOS */
			if (1 == n) {
				t[0] = ParseNum(t[0]);
			} else {
				n = ParseNum(t[0]) * 16;
				t[0] = *(++q);
#ifdef ZOS
				convert_etoa(t);
#endif /* ZOS */
				t[0] = n + ParseNum(t[0]);
			}
		} else {
#ifdef ZOS
				convert_etoa(t);
#endif /* ZOS */
		}
		t[0] ^= s[j % 8];
#ifdef ZOS
		convert_atoe(t);
#endif /* ZOS */
		strncat(result, t, 1);
		q++;
	}

	return result;
}

int FuncExec(void)
{
	char flags[64];
	char *buf = NULL;
	char *ccroot = NULL;
	char *exp_exec = NULL;
	char *userdata = NULL;
	char *dir = NULL;
	char *p, *sptr, *exec, *odir, *tag, *root, *indata, *offline, *q;
	int timeout = DEFAULT_TIMEOUT;
	int result = 0;
	int i, timeout_occurred, preparse;
	size_t userdata_size;
	size_t len = 0;
	size_t cat_count = 0;

	/* Handle the request parameters */
	if (NULL == (exec = request_get(TAG_EXEC))) {
		send_missing(CMD_EXEC, TAG_EXEC);
		return 1;
	}
	if (NULL == (odir = request_get(TAG_DIR))) {
		request_put(TAG_DIR,"");
		odir = request_get(TAG_DIR);
	}
	if (NULL == (tag = request_get(TAG_TAG))) {
		request_put(TAG_TAG,"");
		tag = request_get(TAG_TAG);
	}
	if (NULL == (root = request_get(TAG_ROOT))) {
		request_put(TAG_ROOT,"");
		root = request_get(TAG_ROOT);
	}
	if (NULL != (p = request_get(TAG_TIMEOUT))) {
		timeout = atoi(p);
		if (timeout < 0) { timeout = DEFAULT_TIMEOUT; }
	}
	verbose = !request_has(TAG_SILENT);
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);
	indata = request_get(TAG_STDIN);
	offline = request_get(TAG_OFFLINE);
	if (offline) {

	}

	StoreUserData(userdata, userdata_size);

	/* From this point forward, we are allocating dynamic data structures
	 * that need to be cleaned back up when we're done.
	 */
	if (0 != platform_map_drives(defMapStr, mapStr)) {
		result = 2;
		send_msg("MAP", "MapFail", NULL);
		goto done;
	}

	/*  build the working directory path */
	CleanPath(root);
	CleanPath(tag);
	CleanPath(odir);
	root = env_expand_var(root, 0);
	dir = env_expand_var(odir, 0);
	CleanPath(dir);
	if (0 != strcmp(dir, odir)) {
		char *before = xlate_build_path(root, tag, odir);
		char *after = xlate_build_path(root, tag, dir);
		send_msg("EXEC", "Xlate", "ss", before, after);
		free(after);
		free(before);
	}

	ClobberPasswords();

	if (NULL != (p = platform_getenv("_TIMEOUT"))) {
		timeout = atoi(p);
		if (timeout < 0) {
			timeout = DEFAULT_TIMEOUT;
		} else {
			send_msg("SET", "EnvTimeoutOverride", "i", timeout);
		}
	}



	/* we don't prefix the clearcase view root here because this path */
	/* will be used under a setview shell which will make the path valid */
	/* we just need to postpone validating the directory until later */
	/* we'll go ahead and build the ccroot for the internals, however */
	ccroot = NULL;
	if (clearcase_view) {
		ccroot = malloc(BFBUFSIZ);
		if (BFCONF_EXISTS(cc_suppress_server_root)) {
			format_message(ccroot, "%s/%s/%s/%s",
					this_view(), clearcase_view, tag, dir);
		} else {
			format_message(ccroot, "%s/%s/%s/%s/%s",
					this_view(), clearcase_view, root, tag, dir);
		}
		CleanPath(ccroot);
	}

	buf = malloc(BFBUFSIZ);
	format_message(buf, "%s/%s/%s", root, tag, dir);
	CleanPath(buf);

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(timeout);
	CCMountVobs(timeout);

	/*  make sure we can cd to it. */
	errno = 0;
	if (ccroot) {
		if (platform_chdir(ccroot) != 0) {
			result = 3;
			send_msg("EXEC", "ChdirFail", "si", ccroot, errno);
			goto done;
		}
	} else {
		if (platform_chdir(buf) != 0) {
			result = 4;
			send_msg("EXEC", "ChdirFail", "si", buf, errno);
			goto done;
		}
	}

	result = log_open();
	if (result != 0 && BFCONF_EXISTS(fail_if_log_file_fails)) { goto done; }
	env_dump(0);

	preparse = !BFCONF_EXISTS(no_preparse_command);
	if (preparse) {
		p = platform_getenv("_NO_PREPARSE_COMMAND");
		if (p) {
			platform_freeenv(p);
			preparse = 0;
		}
	}

	timeout_occurred = 0;

	/* see if exec is an internal command */
	/* we pass the unexpanded version of the command string */
	for (i=0; NULL != internals[i].cmd; ++i) {
		len = strlen(internals[i].cmd);
		if (0 == strncmp(internals[i].cmd, exec, len)) {
			switch (exec[len]) {
				case '\0':
				case '\t':
				case '\n':
				case '\r':
				case ' ':
					flags[0] = '\0';
					break;
				case '/': {
					size_t flen;
					p = &exec[++len];
					flen = strcspn(p, " \t\n\r");
					len += flen;
					if (flen >= 64) {
						memcpy(flags, p, 63);
						flags[63] = '\0';
					} else {
						memcpy(flags, p, flen);
						flags[flen] = '\0';
					}
					break;
				}
				default:
					continue;
			}
			break;
		}
	}

	if (NULL != internals[i].cmd) {
		if (preparse) {
			if (NULL != strchr(flags, 'x')) {
				send_msg("EXEC", "ExecPreparse", NULL);
				exp_exec = env_expand_hidden(exec, 1);
				BFTRACE("EXEC built-in / preparse (ignoring escapes)");
			} else {
				send_msg("EXEC", "ExecPreparse", NULL);
				exp_exec = env_expand_hidden(exec, 0);
				BFTRACE("EXEC built-in / preparse");
			}
		} else {
			exp_exec = strdup(exec);
			BFTRACE("EXEC built-in / no preparse");
		}

		send_script(buf, exp_exec, verbose);
		p = &exp_exec[len];
		while (isspace( (int)(*p) )) { ++p; }
		result = (internals[i].func)(p, timeout, &timeout_occurred, flags);
	} else {
		if (preparse) {
			send_msg("EXEC", "ExecPreparse", NULL);
			exp_exec = env_expand_hidden(exec, 0);
			BFTRACE("EXEC external / preparse");
		} else {
			exp_exec = strdup(exec);
			BFTRACE("EXEC external / no preparse");
		}

		/* UNIX systems need to allocate a pseudo-terminal, so we delay
		 * the output of the start message until inside platform_spawn()
		 * on those platforms.
		 */
#ifdef WIN32
		send_script(buf, exp_exec, verbose);
#endif

		q = strdup(exp_exec);
		i = strlen(exp_exec);
		exec = (char *)calloc(i, sizeof(char));
		i = strlen(FORGE);
		p = (char *)strstr(exp_exec, FORGE);
		while (NULL != p) {
			p += i;
			odir = (char *)calloc(1, BFBUFSIZ);
			strncpy(odir, p, (BFBUFSIZ - 1));
			offline = (char *)calloc(1, BFBUFSIZ);
			strncpy(offline, FORGE, (BFBUFSIZ - 1));
			p = (char *)strstr(p, FORGE);
			if (NULL != p) {
				p += i + 8;
				preparse = strlen(odir) - strlen(p);
				*(odir + preparse) = 0;
				sptr = Parse(odir);
				cat_count = (strlen(odir) < BFBUFSIZ) ? strlen(odir) : BFBUFSIZ;
				strncat(offline, odir, cat_count);
				q = BFStrRep(q, offline, sptr);
				cat_count = (strlen(sptr) < BFBUFSIZ) ? strlen(sptr) : BFBUFSIZ;
				strncat(exec, sptr, cat_count);
				strncat(exec, FORGE, i);
				p = (char *)strstr(q, FORGE);
			}
			if (NULL != offline) {
				free(offline);
			}
			if (NULL != odir) {
				free(odir);
			}
		}

		if( NULL !=bfconf_get("bfagent_platform_env_debug", NULL) ) {
			activity_write(BFHERE "spawn command[%s] and flag[%s]", q, exec);
        }
		result = platform_spawn(
				buf,
				q,
				timeout,
				indata,
				sourcelist,
				1,
				&timeout_occurred,
				exec,
				(NULL != trapCommand && '\0' != *trapCommand) ? 0 : 1);

        if (NULL != exec) {
		    free(exec);
	    }

	    if (NULL != q) {
		    free(q);
	    }
	}

	

	/* allow other codes to be treated as a pass */
	if (result && NULL != ( p = platform_getenv("_EXITCODE_MAP") )) {
		char *msg = strdup(p);
		BFTRACE("EXEC exit code map");
		sptr = NULL;
		p = bf_next_token(msg, ",:; ", &sptr);
		while (p && result) {
			i = atoi(p);
			if (result == i) {
				send_msg("EXEC", "ExecExitMap", "i", result);
				result = 0;
			}
			p = bf_next_token(NULL, ",:; ", &sptr);
		}
		free(msg);
	}

	if (result && trapCommand && strlen(trapCommand)) {
		BFTRACE("EXEC trap");
		send_msg("EXEC", "ExecTrap", "is", result, trapCommand);
		trapmode = 1;
		platform_spawn(buf, trapCommand, 60, NULL, NULL, 0, &timeout_occurred, NULL, 1);
		trapmode = 0;
	}

/*	p = bf_expose_special_chars(exp_exec); */
	if (timeout_occurred) {
		BFTRACE("EXEC timeout");
		send_msg("TMO", "ExecTimeout", "i", timeout_occurred);
	} else {
		BFTRACE("EXEC ok");
		send_msg("EXEC", "ExecOk", "ss", buf, platform_hostname());
	}
/*	free(p); */

done:
	if (result != 0) { env_dump(result); }
	log_close();
	/* platform_unmap_drives(0); */
	FreeUserData();
	free(exp_exec);
	free(ccroot);
	free(dir);
	free(root);
	free(buf);
	return result;
}

