/*
Copyright (C)2003-2013, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/* Low-level I/O abstraction for hiding whether or not SSL is being
 * used, transcoding to/from UTF-8 for the protocol, hiding the fact
 * that a file descriptor and a Socket are not interchangable on
 * Windows, and so on.
 */

/* Activity logging for this file is suppressed unless IO_DEBUG
 * is defined during compilation.
 */
#ifndef IO_DEBUG
#  undef activity_file
#  define activity_file NULL
#endif

/* #define IO_DEBUG_SSL_INSECURE 1 */

/* GLOBAL STATE
 *
 * - ssl_state
 * Tracks whether or not SSL is being used and whether or not we did a cert
 * verification to confirm the identity of the client (engine) that connected
 * to us.
 */

ssl_state_t ssl_state = SS_OK;

/* Flag to show the system ssl related sensitive information in the activity log flag 
   system_ssl_debug = bfconf_get("bfagent_system_ssl_debug", NULL);
 */
static	char *          system_ssl_debug  = NULL;

#ifdef HAVE_SYSTEM_SSL

#define hRemote ofd 

#undef D
#undef HAVE_SSL      /* not use openssl   */

#ifdef SYSTEM_SSL_DEBUG
#  define D(x) x
# else
#  define D(x) /* ignore */
#endif /* SYSTEM_SSL_DEBUG/! */

extern int win_sock;
extern int ofd;
extern int ifd;

extern int encryptmode;

void *ssl = NULL;

static gsk_handle env_handle;
static gsk_handle soc_handle;

static GSK_ENUM_VALUE sslv2, sslv3, tlsv10, tlsv11, tlsv12;


int secureSocRecv(int fd, void *data, int len, char *user_data) {
	return( recv( fd, data, len,0 ));
}
int secureSocSend(int fd, void *data, int len, char *user_data) {
	return(send( fd, data, len,0 ));
}


int gsk_ssl_read(void *buf, int num)
{
	int status;
	int len ;
	char *buffer = (char *)buf;
	/* int size = strlen(buffer); */
	int size = (BFBUFSIZ << 1);
		
	status = gsk_secure_socket_read(soc_handle, buffer, size, &len);
	if(GSK_OK != status){
		D(activity_write(BFHERE "gsk_ssl_read ERROR : %d", status));
		return -1;
	}
	
	D(activity_write(BFHERE "gsk_ssl_read: %d bytes of %d\n", len , size)); 
	D(activity_write(BFHERE "gsk_ssl_read-buffer: %s\n", buffer ));
	
	return len;
}

int gsk_ssl_write(const void *buf, int num)
{
	int status;
	int len ;
	char *buffer = (char *)buf;
	int size = strlen(buffer);	
	
	status = gsk_secure_socket_write(soc_handle, buffer, size, &len);
	if(GSK_OK != status){
		D(activity_write(BFHERE "gsk_ssl_write ERROR : %d", status));
		return -1;
	}
	
	D(activity_write(BFHERE "gsk_ssl_write: %d bytes of %d\n", len , size)); 
	D(activity_write(BFHERE "gsk_ssl_write-buffer: %s\n", buffer)); 
	
	return len;
}


void check_return_set(int rc,char* variable) {
  if (rc != 0) {
    printf("SERVER: Error setting variable %s.\n",variable);
    printf("    Return Code: %d\n",rc);
    printf("    Message: %s\n",gsk_strerror(rc));
    gsk_environment_close(&env_handle);
    exit(0);

  }
}

void show_cert (char * cert_type, const gsk_cert_data_elem * cert_ptr,int cert_elems) 
{
  const gsk_cert_data_elem * my_cert;  
  int cert_elem;                
  my_cert = cert_ptr;            
  D(activity_write(BFHERE "Information about %s certificate:\n", cert_type));
  
  
  for (cert_elem = 0; cert_elem < cert_elems; cert_elem++) {
    switch (my_cert[cert_elem].cert_data_id) {
      case CERT_BODY_DER :
	    D(activity_write(BFHERE "CERT_BODY_DER length is: %d.\n", my_cert[cert_elem].cert_data_l));
        break;
      case CERT_BODY_BASE64 :
	    D(activity_write(BFHERE "CERT_BODY_BASE64 length is: %d.\n", my_cert[cert_elem].cert_data_l));
        break;
      case CERT_SERIAL_NUMBER :
	    D(activity_write(BFHERE "CERT_SERIAL_NUMBER is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_COMMON_NAME :
	    D(activity_write(BFHERE "CERT_COMMON_NAME is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_LOCALITY :
	    D(activity_write(BFHERE "CERT_LOCALITY is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_STATE_OR_PROVINCE :
	    D(activity_write(BFHERE "CERT_STATE_OR_PROVINCE is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_COUNTRY :
	    D(activity_write(BFHERE "CERT_COUNTRY is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ORG :
	    D(activity_write(BFHERE "CERT_ORG is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ORG_UNIT :
	    D(activity_write(BFHERE "CERT_ORG_UNIT is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_DN_PRINTABLE :
	    D(activity_write(BFHERE "CERT_DN_PRINTABLE is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_DN_DER :
	    D(activity_write(BFHERE "CERT_DN_DER length is: %d.\n", my_cert[cert_elem].cert_data_l));
        break;
      case CERT_POSTAL_CODE :
	    D(activity_write(BFHERE "CERT_POSTAL_CODE is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_EMAIL :
	    D(activity_write(BFHERE "CERT_EMAIL is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_COMMON_NAME :
	    D(activity_write(BFHERE "CERT_ISSUER_COMMON_NAME is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_LOCALITY :
	   D(activity_write(BFHERE "CERT_ISSUER_LOCALITY is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_STATE_OR_PROVINCE :
	    D(activity_write(BFHERE "CERT_ISSUER_STATE_OR_PROVINCE is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_COUNTRY :
	    D(activity_write(BFHERE "CERT_ISSUER_COUNTRY is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_ORG :
	    D(activity_write(BFHERE "CERT_ISSUER_ORG is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_ORG_UNIT :
	    D(activity_write(BFHERE "CERT_ISSUER_ORG_UNIT is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_DN_PRINTABLE :
	    D(activity_write(BFHERE "CERT_ISSUER_DN_PRINTABLE is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_DN_DER :
	    D(activity_write(BFHERE "CERT_ISSUER_DN_DER length is: %d.\n", my_cert[cert_elem].cert_data_l));
        break;
      case CERT_ISSUER_POSTAL_CODE :
	    D(activity_write(BFHERE "CERT_ISSUER_POSTAL_CODE is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      case CERT_ISSUER_EMAIL :
	    D(activity_write(BFHERE "CERT_ISSUER_EMAIL is: %s.\n", my_cert[cert_elem].cert_data_p));
        break;
      default:
	    D(activity_write(BFHERE "Unrecognized cert data type: %d.\n", my_cert[cert_elem].cert_data_id));
        break;
    }                  
  }                       
}                            


void get_peer_certificate(void)
{
	int rc;
	gsk_cert_data_elem * local_cert ; 
	int  local_cert_elem_ct;          
    gsk_cert_data_elem * partner_cert ; 
    int  partner_cert_elem_ct;  
	
	rc = gsk_attribute_get_cert_info (soc_handle,GSK_LOCAL_CERT_INFO, &local_cert, &local_cert_elem_ct);
	if (rc != GSK_OK) {    
		activity_write(BFHERE "Error %d from 'gsk_attribute_get_cert_info(LOCAL)'\n", rc);		
	}else if (local_cert_elem_ct > 0) {  
        show_cert("LOCAL", local_cert, local_cert_elem_ct);
        gsk_free_cert_data((gsk_cert_data_elem*)local_cert, local_cert_elem_ct);
    }
	
	rc = gsk_attribute_get_cert_info (soc_handle, GSK_PARTNER_CERT_INFO, &partner_cert, &partner_cert_elem_ct);
    if (rc != GSK_OK) {                
		activity_write(BFHERE "Error %d from 'gsk_attribute_get_cert_info(PARTNER)'\n", rc);	
    }else if (partner_cert_elem_ct > 0) {  
        show_cert("PARTNER", partner_cert, partner_cert_elem_ct);
        gsk_free_cert_data((gsk_cert_data_elem*)partner_cert, partner_cert_elem_ct);
    }
	
}


static ssl_state_t init_system_ssl_environment(void){
	int rc, rcset;
	char * ssl_key_location, *sptr, *ssl_saf_key_ring;
	char * ssl_kdb_password, *ssl_password_encrypt, *decode_password;
	char * ssl_protocol;
	char * ssl_key_ring_label;
	char * temp = NULL;
	int saf = 0;
	
	ssl_key_location= bfconf_get("gsk_ssl_key_location", NULL);
	/* activity_write(BFHERE "strlen ssl_key_location : %d\n", strlen(ssl_key_location)); */
	if (NULL == ssl_key_location) {
		BFTRACE("SSL gsk_ssl_key_location=(null) -> SSL off");
		return SS_OK;
	} else {
		if (NULL != activity_file) {
			activity_write(BFHERE "ssl_key_location : %s\n", ssl_key_location);
		}
		
		temp=malloc(strlen(ssl_key_location)+1);
		memcpy(temp, ssl_key_location, strlen(ssl_key_location)+1);
		
		sptr = NULL;
		ssl_saf_key_ring = bf_next_token(temp, " \t\r\n", &sptr);
		
		if ( NULL != sptr) {
			if (NULL != activity_file) {
				activity_write(BFHERE "sptr : %s\n", sptr);
			}
			/* sptr = LowerString(sptr);	
			activity_write(BFHERE "strlen sptr : %d\n", strlen(sptr));
			*/
			if(strcmp(LowerString(sptr), "saf") == 0){
				/* SAF key ring used */
				saf = 1;
				if (NULL != activity_file) {
					activity_write(BFHERE "saf_key_ring : %s\n", ssl_saf_key_ring);
				}
			}
		}
	}
	
	if( !saf) {

		ssl_kdb_password=bfconf_get("gsk_ssl_kdb_password", NULL);
		if (NULL == ssl_kdb_password) {
			BFTRACE("SSL ssl_kdb_password=(null) -> SSL off");
			return SS_OK;
		}
		activity_write(BFHERE "gsk_ssl_kdb_password : %s\n", ssl_kdb_password);
	
		ssl_password_encrypt=bfconf_get("gsk_password_encrypt", NULL);
		if (NULL != ssl_password_encrypt && 0 == strcmp(LowerString(ssl_password_encrypt), "true")){
		
			decrypt_pwd(&decode_password, ssl_kdb_password, 0);
			
			system_ssl_debug = bfconf_get("bfagent_system_ssl_debug", NULL);
			if(NULL == system_ssl_debug ) {
			   BFTRACE("decode_password :******");
			}else{
			   activity_write(BFHERE "decode_password : %s\n", decode_password);
			}
		}
	}
	
	ssl_key_ring_label= bfconf_get("gsk_keyring_label", NULL);
	if (NULL == ssl_key_ring_label) {
		BFTRACE("SSL ssl_key_ring_label=(null) -> SSL off");
		return SS_OK;
	}
	
	ssl_protocol = bfconf_get("gsk_ssl_protocol", "tlsv10");
	ssl_protocol = LowerString(ssl_protocol);
	if (strcmp(ssl_protocol, "sslv2") == 0) {
          sslv2 = GSK_PROTOCOL_SSLV2_ON;
          sslv3 = GSK_PROTOCOL_SSLV3_OFF;
          tlsv10 = GSK_PROTOCOL_TLSV1_OFF;
          tlsv11 = GSK_PROTOCOL_TLSV1_1_OFF;
          tlsv12 = GSK_PROTOCOL_TLSV1_2_OFF;
	}else if(strcmp(ssl_protocol, "sslv3") == 0){
          sslv2 = GSK_PROTOCOL_SSLV2_OFF;
          sslv3 = GSK_PROTOCOL_SSLV3_ON;
          tlsv10 = GSK_PROTOCOL_TLSV1_OFF;
          tlsv11 = GSK_PROTOCOL_TLSV1_1_OFF;
          tlsv12 = GSK_PROTOCOL_TLSV1_2_OFF;
	}else if(strcmp(ssl_protocol, "tlsv10") == 0){
          sslv2 = GSK_PROTOCOL_SSLV2_OFF;
          sslv3 = GSK_PROTOCOL_SSLV3_OFF;
          tlsv10 = GSK_PROTOCOL_TLSV1_ON;
          tlsv11 = GSK_PROTOCOL_TLSV1_1_OFF;	
          tlsv12 = GSK_PROTOCOL_TLSV1_2_OFF;
	}else if(strcmp(ssl_protocol, "tlsv11") == 0){
          sslv2 = GSK_PROTOCOL_SSLV2_OFF;
          sslv3 = GSK_PROTOCOL_SSLV3_OFF;
          tlsv10 = GSK_PROTOCOL_TLSV1_OFF;
          tlsv11 = GSK_PROTOCOL_TLSV1_1_ON;	
          tlsv12 = GSK_PROTOCOL_TLSV1_2_OFF;
	}else if(strcmp(ssl_protocol, "tlsv12") == 0){
          sslv2 = GSK_PROTOCOL_SSLV2_OFF;
          sslv3 = GSK_PROTOCOL_SSLV3_OFF;
          tlsv10 = GSK_PROTOCOL_TLSV1_OFF;
          tlsv11 = GSK_PROTOCOL_TLSV1_1_OFF;
          tlsv12 = GSK_PROTOCOL_TLSV1_2_ON;
	}else if(strcmp(ssl_protocol, "all") == 0){
		  sslv2 = GSK_PROTOCOL_SSLV2_ON;
          sslv3 = GSK_PROTOCOL_SSLV3_ON;
          tlsv10 = GSK_PROTOCOL_TLSV1_ON;
          tlsv11 = GSK_PROTOCOL_TLSV1_1_ON;	
          tlsv12 = GSK_PROTOCOL_TLSV1_2_ON;
	}else {
		  BFTRACE("invalid value specified for protocol");
		  return SS_OK;
    }
	
	rc = gsk_environment_open(&env_handle);
    if (rc) {
		activity_write(BFHERE "Error in gsk_environment_open rc: %d\n", rc);
		gsk_environment_close(&env_handle);
		return SS_FAILED;
	} 
	
	if (!saf){
		rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_FILE,ssl_key_location,0);
		activity_write(BFHERE "SSL ssl_key_location=%s", ssl_key_location);
	} else {
		rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_FILE,ssl_saf_key_ring,0);
		activity_write(BFHERE "SSL ssl_saf_key_ring=%s", ssl_saf_key_ring);
	}
	
	
	if (!saf) {
		if (NULL != ssl_password_encrypt && 0 == strcmp(LowerString(ssl_password_encrypt), "true")){
			rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_PW,decode_password,0);
			
			system_ssl_debug = bfconf_get("bfagent_system_ssl_debug", NULL);
			if(NULL == system_ssl_debug ) {
			   BFTRACE("SSL decode_password=******");
			}else{
			  activity_write(BFHERE "SSL decode_password=%s", decode_password);
			}

		} else {
			rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_PW,ssl_kdb_password,0);
			
			if(NULL == system_ssl_debug ) {
			   BFTRACE("SSL ssl_kdb_password=******");
			}else{
			  activity_write(BFHERE "SSL ssl_kdb_password=%s", ssl_kdb_password);
			}
			
		}
		
		/*
		if(NULL == ssl_password_encrypt){
			rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_PW,ssl_kdb_password,0);
			activity_write(BFHERE "SSL ssl_kdb_password=%s", ssl_kdb_password);
		}else{
			rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_PW,decode_password,0);
			activity_write(BFHERE "SSL decode_password=%s", decode_password);
		}
		*/
	} else {
		rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_PW,NULL,0);
		BFTRACE("SAF keyring: set GSK_KEYRING_PW as NULL");
	}
	
	
	
	rcset = gsk_attribute_set_enum(env_handle,GSK_PROTOCOL_SSLV2,sslv2);
    rcset = gsk_attribute_set_enum(env_handle,GSK_PROTOCOL_SSLV3,sslv3);
	rcset = gsk_attribute_set_enum(env_handle,GSK_PROTOCOL_TLSV1,tlsv10);
    rcset = gsk_attribute_set_enum(env_handle,GSK_PROTOCOL_TLSV1_1,tlsv11);
	rcset = gsk_attribute_set_enum(env_handle, GSK_PROTOCOL_TLSV1_2, tlsv12);
	activity_write(BFHERE "Server: Setting protocol to %s.\n", ssl_protocol);
	
	rcset = gsk_attribute_set_buffer(env_handle,GSK_KEYRING_LABEL,ssl_key_ring_label, 0);
	activity_write(BFHERE "SSL ssl_key_ring_label=%s", ssl_key_ring_label);
	
	
	/* initialize environment */
	rc = gsk_environment_init(env_handle);
	
	if(GSK_OK != rc){
		activity_write(BFHERE "SYSTEM SSL env init failed,failed code rc: %d\n", rc);
		gsk_environment_close(&env_handle);
		return SS_FAILED;
	}
	
	BFTRACE("SYSTEM SSL env init ok");
	
	if(NULL != temp){
		free(temp), temp = NULL;
	}
	
	return SS_CERTIFIED;
}

void stop_SSL(void) {
	send_goodbye();
	if (ssl) {
		gsk_secure_socket_close(&soc_handle);
		gsk_environment_close(&env_handle);
		ssl = NULL;
	}

	close(0);
	close(1);
	close(2);

	if (ifd > 2) { close(ifd); }
	if (ofd > 2 && ifd != ofd) { close(ofd); }
}

	
void report_SSL(void) {
	char *line;
	gsk_cipher_suites *cipher;

	if (!ssl) {
		send_msg("SSL", "SSLPlainText", NULL);
		return;
	}
	
	if (ssl_state != SS_CERTIFIED) {
		send_msg("SSL", "SSLUntrustedCert", NULL);
	}
	
	/* Get certificates (if available) */
	get_peer_certificate();
	
}
	
void *start_SSL(void) {
	int rc, rcset;
	char *cipher_specs_V2;
	char *cipher_specs_V3;
	char *clientAuth;
	gsk_iocallback local_io = {secureSocRecv, secureSocSend, NULL, NULL, NULL, NULL};
	gsk_cipher_suites *cipher;
	
	BFTRACE("In start_SSL : initialize system ssl environment");
	ssl_state = init_system_ssl_environment();
	
	if (SS_FAILED == ssl_state) { return NULL; }
	
	/* You cannot send a hello in the middle of an SSL handshake.  Defer
	   the SSL hello until the end.  Send the TCP hello now, if the state 
	   is SS_OK, which means we aren't using SSL.
	 */
	if (SS_OK == ssl_state) {
		send_hello(CODE_HELLO);
		return NULL;
	}
	
	BFTRACE("Calling secure socket open ");
	rc = gsk_secure_socket_open( env_handle, &soc_handle );
	if (GSK_OK != rc) {
		BFTRACE("SSLERROR: failed to open secure socket ");
		gsk_secure_socket_close(&soc_handle);
		gsk_environment_close(&env_handle);
		ssl_state = SS_FAILED;
		
		return NULL;
    }
	
	/* hRemote is the duplicate file descriptor from the dispatch thread. */
	rcset = gsk_attribute_set_numeric_value(soc_handle,GSK_FD, hRemote);
	/* check_return_set(rcset,"GSK_FD"); */
		
	clientAuth= bfconf_get("gsk_ssl_client_authentication", NULL);
	if (NULL != clientAuth && 0 == strcmp(LowerString(clientAuth), "true")){
		rcset = gsk_attribute_set_enum(soc_handle,GSK_SESSION_TYPE,GSK_SERVER_SESSION_WITH_CL_AUTH);
		activity_write(BFHERE "SSL GSK_SESSION_TYPE=GSK_SERVER_SESSION_WITH_CL_AUTH", rcset);
	} else {
		rcset = gsk_attribute_set_enum(soc_handle,GSK_SESSION_TYPE,GSK_SERVER_SESSION);
		activity_write(BFHERE "SSL GSK_SESSION_TYPE=GSK_SERVER_SESSION", rcset);
	}
	
	
	cipher_specs_V2=bfconf_get("gsk_ssl_cipher_v2", "631");
	rcset = gsk_attribute_set_buffer(soc_handle,GSK_V2_CIPHER_SPECS,cipher_specs_V2,0);
	/* check_return_set(rcset,"GSK_V2_CIPHER_SPECS"); */
	activity_write(BFHERE "SSL cipher_specs_V2=%s", cipher_specs_V2);
	
	cipher_specs_V3=bfconf_get("gsk_ssl_cipher_v3", "090201");
	rcset = gsk_attribute_set_buffer(soc_handle,GSK_V3_CIPHER_SPECS,cipher_specs_V3,0);
	/* check_return_set(rcset,"GSK_V3_CIPHER_SPECS"); */
	activity_write(BFHERE "SSL cipher_specs_V3=%s", cipher_specs_V3);
	
	rc = gsk_attribute_set_callback(soc_handle, GSK_IO_CALLBACK, &local_io);
	
	BFTRACE("Calling gsk_secure_socket_init() to start ssl handshake with client.");
	/*  real ssl handshake with client*/
	rc = gsk_secure_socket_init( soc_handle );	
	activity_write(BFHERE "SSL_accept rc=%d (0 is good, >=1 is handshake failure or fatal)", rc);
	
	if(GSK_OK != rc){
		BFTRACE(" SSLErrorHandshake: failed to Call gsk_secure_socket_init() to.");
		gsk_secure_socket_close(&soc_handle);
		gsk_environment_close(&env_handle);
		
		ssl_state = SS_FAILED;
		return NULL;
	}
	
	BFTRACE(" SSLHandshake: sucessfull initialize secure socket by calling gsk_secure_socket_init() .");
	ssl = &soc_handle;
	
	/*
	gsk_get_cipher_suites(cipher);
	BFTRACE(" gsk_get_cipher_suites(cipher) .");

	if (NULL != cipher) {
		activity_write(BFHERE "Cipher v2 chosen: %s", cipher->v2_cipher_suites);
		activity_write(BFHERE "Cipher v3 chosen: %s", cipher->v3_cipher_suites);
	}
	*/
	
	get_peer_certificate();
	BFTRACE("Sucessfully to get certificates.");
	
	switch (ssl_state) {
		case SS_OK:
			BFTRACE("ssl_state = SS_OK");
			send_hello(CODE_HELLO);
			return NULL;
		case SS_SECURE:
			BFTRACE("ssl_state = SS_SECURE");
			send_hello(CODE_SECURE_HELLO);
			break;
		case SS_CERTIFIED:
			BFTRACE("ssl_state = SS_CERTIFIED");
			send_hello(CODE_CERTIFIED_HELLO);
			break;
		default:
			return NULL;
	}
	
	
	return ssl;
}	

#define io_read(buf, len) (ssl ? gsk_ssl_read(buf, len) : platform_protocol_read(buf, len))
#define io_write(buf, len) (ssl ? gsk_ssl_write(buf, len) : platform_protocol_write(buf, len))


#else /* ! HAVE_SYSTEM_SSL */

#ifdef HAVE_SSL 
/*================================================================
 * Compiled with OpenSSL
 *----------------------------------------------------------------*/

#ifdef WIN32
extern SOCKET hRemote;
# else
#  define hRemote ofd 
#endif
extern int win_sock;
extern int ofd;
extern int ifd;

SSL_CTX *ctx = NULL;
void *ssl = NULL;

/* When OpenSSL finds an X.509 cert and wants to inform us whether or not
 * it thinks we should trust that cert, it calls this function.  We simply
 * trust its opinion, here.
 */
static int verify_SSL_callback(int flag, X509_STORE_CTX *store) {
	int 	err;

	if (!flag) {
		err = X509_STORE_CTX_get_error(store);
		activity_write(BFHERE "SSL AUTH FAIL Certificate error: %d: %s",
				err, X509_verify_cert_error_string(err));
		send_msg("SSL", "SSLErrorUnverifiedCert", "is",
				err, X509_verify_cert_error_string(err));
		return 0;
	}
	return 1;
}

/* Convenience method to make sure that we provide as much information
 * as we can for troubleshooting SSL problems.
 */
static void report_ssl_error(const char *file, int line, const char *reason, int rc) {
	int saved_errno = errno;
	char err_buf[1024];
	unsigned long err = ERR_get_error();
	int sslerr;

	if (NULL != ssl) { 
		sslerr = SSL_get_error(ssl, rc);
	} else {
		sslerr = 0;
	}

	*err_buf = '\0';  /* Just in case... */
	ERR_error_string_n(err, err_buf, sizeof(err_buf));

	activity_write(file, line, "SSL FAIL");
	activity_write(BFHERE "- reason : %s", reason);
	activity_write(BFHERE "- err_buf: %s", err_buf);
	activity_write(BFHERE "- err    : %08lX", err);
	activity_write(BFHERE "- sslerr : %d", sslerr);
	activity_write(BFHERE "- errno  : %d", saved_errno);
	activity_write(BFHERE "- rc     : %d", rc);

	send_msg("SSL", reason, "s", err_buf);
	errno = saved_errno;
}

/* Retrieves the passphrase for the private key from bfagent.conf */
static char *init_ssl_passphrase(const char *tag, const int sha_back) {
	char *pass = getPasswordFromAgentConfig(tag, sha_back);
	if (NULL != pass) {
#ifdef IO_DEBUG_SSL_INSECURE
		activity_write(BFHERE "SSL PASS %s=%s", tag, pass);
# else /* !IO_DEBUG_SSL_INSECURE */
		activity_write(BFHERE "SSL PASS %s=(set)", tag);
#endif /* IO_DEBUG_SSL_INSECURE/! */
	} else {
		pass = strdup("");
		activity_write(BFHERE "SSL PASS %s=(null)", tag);
	}
	SSL_CTX_set_default_passwd_cb_userdata(ctx, pass);
	return pass;
}

static int init_ssl_core(void) {
	/* Per docs, these never fail. */
	SSL_load_error_strings();
	ERR_load_crypto_strings();
	SSL_library_init();
	BFTRACE("SSL CORE ok");
	return 1;
}

static int init_ssl_fips(void) {
	char *fips_enabled;

	/* See if FIPS is enabled early on. */
	fips_enabled = bfconf_get("fips_enabled", NULL);
	if (NULL == fips_enabled) {
		BFTRACE("fips_enabled=false  [default]");
		return 1;
	}

	activity_write(BFHERE "fips_enabled=%s", (NULL != fips_enabled) ? fips_enabled : "(null)");
	if (0 != strcmp("true", fips_enabled)) {
		free(fips_enabled);
		return 1;
	}

#ifdef OPENSSL_FIPS
	/* Check if FIPS already enabled. */
	if (FIPS_mode()) { return 1; }
	if (FIPS_mode_set(1)) { 
		BFTRACE("SSL FIPS enabled");
		return 1;
	}
# else /* !OPENSSL_FIPS */
	BFTRACE("SSL FIPS mode requested, but this OpenSSL library does not support it!");
#endif /* OPENSSL_FIPS/! */
	report_ssl_error(BFHERE "SSLErrorFIPSEnablement", 0);
	return 0;
}



static int init_ssl_protocol(void) {
	char *proto;
#ifdef OPENSSL_FIPS
	if (FIPS_mode()) {
		BFTRACE("SSL PROTO TLSv1 [FIPS]");
		proto = strdup("TLSv1");
	} else
#endif /* OPENSSL_FIPS */
	{
		proto = bfconf_get("ssl_protocol", "TLSv1");
		activity_write(BFHERE "SSL PROTO %s", proto);
	}

	/* Create the SSL context with the desired protocol. */
	if (0 == strcmp(proto, "TLSv1") || 0 == strcmp(proto, "TLS")) {
		BFTRACE("SSL Creating TLSv1 ctx");
		ctx = SSL_CTX_new(TLSv1_server_method());
	} else if ((0 == strcmp(proto, "TLSv1.1")) || (0 == strcmp(proto, "TLSv1_1")) || (0 == strcmp(proto, "TLSv11"))) {
		BFTRACE("SSL Creating TLSv1.1 ctx");
		ctx = SSL_CTX_new(TLSv1_1_server_method());
	} else if ((0 == strcmp(proto, "TLSv1.2")) || (0 == strcmp(proto, "TLSv1_2")) || (0 == strcmp(proto, "TLSv12"))) {
		BFTRACE("SSL Creating TLSv1.2 ctx");
		ctx = SSL_CTX_new(TLSv1_2_server_method());
	} else {
		BFTRACE("SSL Creating SSLv3 ctx");
		ctx = SSL_CTX_new(SSLv3_server_method());
	}

	if (NULL == ctx) {
		report_ssl_error(BFHERE "SSLErrorNoCtx", 0);
		free(proto);
		return 0;
	}

	SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
	SSL_CTX_set_options(ctx, SSL_OP_TLS_ROLLBACK_BUG);

	free(proto);
	return 1;
}

/* An SSL group will be HIGH, MEDIUM, LOW, or ALL (among other OpenSSL cipher selections). */
static int init_ssl_cipher_group(void) {
	int rc;
	char *cipher_group = bfconf_get("ssl_cipher_override", NULL);

	if (NULL != cipher_group) {
		activity_write(BFHERE "SSL CIPHER %s [Override]", cipher_group);
	} else {
#ifdef OPENSSL_FIPS
		if (FIPS_mode()) {
			BFTRACE("SSL CIPHER HIGH [FIPS]");
			cipher_group = strdup("HIGH");
		} else
#endif /* OPENSSL_FIPS */
		{
			cipher_group = bfconf_get("ssl_cipher_group", "ALL");
			activity_write(BFHERE "SSL CIPHER %s", cipher_group);
		}
	}

	rc = SSL_CTX_set_cipher_list(ctx, cipher_group);
	if (1 != rc) {
		report_ssl_error(BFHERE "SSLErrorInvalidCipher", rc);
		free(cipher_group);
		return 0;
	}

	BFTRACE("SSL CIPHER ok");
	free(cipher_group);
	return 1;
}


static int init_ssl_cert(const char *cert) {
	int rc;
	char *pass;

	/* Load the cert. */
	BFTRACE("SSL CERT load");
	pass = init_ssl_passphrase("ssl_cert_password", 0);

	rc = SSL_CTX_use_certificate_chain_file(ctx, cert);
	if (1 != rc) {
		report_ssl_error(BFHERE "SSLErrorBadCert", rc);
		free(pass);
		return 0;
	}

	BFTRACE("SSL CERT ok");
	free(pass);
	return 1;
}


static int init_ssl_ca_cert(int *certified) {
	STACK_OF(X509_NAME) *cert_names;
	char *ca_loc = NULL;
	char *client_auth = NULL;
	char *pass = NULL;
	int rc = 1;

	*certified = 0;

	/* Load the trust store that contains the signers. */
	ca_loc = bfconf_get("ssl_ca_location", NULL);
	if (NULL == ca_loc) {
		BFTRACE("SSL CA none");
		return 1;
	}
	
	activity_write(BFHERE "SSL CA Location: %s", ca_loc);
	/* Set the password for accessing the CA store, if required. */
	pass = init_ssl_passphrase("ssl_ca_password", 0);

	rc = SSL_CTX_load_verify_locations(ctx, ca_loc, NULL);
	if (1 != rc) {
		report_ssl_error(BFHERE "SSLErrorBadCA", rc);
		goto done;
	}

	/* Check if client_authentication verification is desired. */
	client_auth = bfconf_get("ssl_client_authentication", NULL);
	if (NULL != client_auth && 0 == strcmp("true", client_auth)) {
		BFTRACE("SSL CA certified");
		/* The list of CA certs sent to the client must be set this way. */
		cert_names = SSL_load_client_CA_file(ca_loc);
		if (NULL != cert_names) { SSL_CTX_set_client_CA_list(ctx, cert_names); }
		SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, verify_SSL_callback);
		*certified = 1;
	} else {
		BFTRACE("SSL CA normal");
		SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);
	}

done:
	if (NULL != ca_loc) { free(ca_loc); }
	if (NULL != pass) { free(pass); }
	if (NULL != client_auth) { free(client_auth); }
	if (1 != rc) { return 0; }
	BFTRACE("SSL CA ok");
	return 1;
}

static int init_ssl_private_key(const char *pkey) {
	char *pass;
	int rc;

	BFTRACE("SSL PKey load");
	pass = init_ssl_passphrase("ssl_key_password", 0);

	rc = SSL_CTX_use_PrivateKey_file(ctx, pkey, SSL_FILETYPE_PEM);
	if (1 != rc) {
		BFTRACE("Failed to use private key, try SHA1");
		pass = init_ssl_passphrase("ssl_key_password", 1);
		rc = SSL_CTX_use_PrivateKey_file(ctx, pkey, SSL_FILETYPE_PEM);
	}
	if (1 != rc) {
		report_ssl_error(BFHERE "SSLErrorBadPKeyFile", rc);
		goto done;
	}
	rc = SSL_CTX_check_private_key(ctx);
	if (1 != rc) {
		report_ssl_error(BFHERE "SSLErrorBadPKey", rc);
		goto done;
	}
	BFTRACE("SSL PKey ok");

done:
	free(pass);
	return (1 != rc) ? 0 : 1;
}

static ssl_state_t init_ssl_ctx(void) {
	char *ssl_key_location;
	char *ssl_cert_location;
	int ok;
	int certified = 0;

	BFTRACE("SSL init");

	/* Paranoid... init_ssl_ctx shouldn't get called more than once... */
	if (ctx) {
		BFTRACE("SSL FAIL re-initialized?!");
		return SS_FAILED;
	}

	/* Both the key and the cert have to be provided to use SSL */
	ssl_key_location = bfconf_get("ssl_key_location", NULL);
	if (NULL == ssl_key_location) {
		BFTRACE("SSL ssl_key_location=(null) -> SSL off");
		return SS_OK;
	}
	activity_write(BFHERE "SSL ssl_key_location=%s", ssl_key_location);

	ssl_cert_location = bfconf_get("ssl_cert_location", NULL);
	if (NULL == ssl_cert_location) {
		BFTRACE("SSL ssl_cert_location=(null) -> SSL off");
		free(ssl_key_location);
		return SS_OK;
	}
	activity_write(BFHERE "SSL ssl_cert_location=%s", ssl_cert_location);
	
	ok = init_ssl_core() &&
		init_ssl_fips() &&
		init_ssl_protocol() &&
		init_ssl_cipher_group() &&
		init_ssl_cert(ssl_cert_location) &&
		init_ssl_ca_cert(&certified) &&
		init_ssl_private_key(ssl_key_location);

	if (NULL != ssl_key_location) { free(ssl_key_location); }
	if (NULL != ssl_cert_location) { free(ssl_cert_location); }

	if (!ok) {
		if (NULL != ctx) {
			SSL_CTX_free(ctx);
			ctx = NULL;
		}
		return SS_FAILED;
	}
	
	BFTRACE("SSL init ok");
	return certified ? SS_CERTIFIED : SS_SECURE;
}

void stop_SSL(void) {
	send_goodbye();
	if (ssl) {
		SSL_shutdown(ssl);
		SSL_free(ssl);
		ssl = NULL;
	}

	close(0);
	close(1);
	close(2);

#ifdef WIN32
	if (win_sock) {
		shutdown(ifd, 0);
		closesocket(ifd);
		WSACleanup();
	}
#else /* !WIN32 */
	if (ifd > 2) { close(ifd); }
	if (ofd > 2 && ifd != ofd) { close(ofd); }
#endif /* WIN32/! */
}


void *start_SSL(void) {
	int saved_errno;
	int rc=0;
	X509* client_cert;
	const char *sbuf;
	
	BFTRACE("In start_SSL");

	/* Specify the context configuration options */
	ssl_state = init_ssl_ctx();
	if (SS_FAILED == ssl_state) { return NULL; }

	/* You cannot send a hello in the middle of an SSL handshake.  Defer
	   the SSL hello until the end.  Send the TCP hello now, if the state 
	   is SS_OK, which means we aren't using SSL.
	 */
	if (SS_OK == ssl_state) {
		send_hello(CODE_HELLO);
		return NULL;
	}

	BFTRACE("Calling SSL_new");
	/* Create a new SSL context */
	ssl = SSL_new(ctx);
	if (NULL == ssl) {
		report_ssl_error(BFHERE "SSLErrorNoSSL", rc);
		ssl_state = SS_FAILED;
		return NULL;
	}

	/* hRemote is the duplicate file descriptor from the dispatch thread. */
	SSL_set_fd(ssl, hRemote);
	SSL_set_accept_state(ssl);

	BFTRACE("Calling SSL_accept.");
	rc = SSL_accept(ssl);
	saved_errno = errno;
	activity_write(BFHERE "SSL_accept rc=%d (1 is good, 0 is handshake failure, <0 is fatal)", rc);
	errno = saved_errno;

	if (rc <= 0) {
		report_ssl_error(BFHERE "SSLErrorHandshake", rc);
		ssl_state = SS_FAILED;
		SSL_free(ssl);
		ssl = NULL;
		return NULL;
	}

	sbuf = SSL_get_cipher_name(ssl);
	if (NULL != sbuf) {
		activity_write(BFHERE "Cipher chosen: %s", sbuf);
	}

	client_cert = SSL_get_peer_certificate(ssl);
	if (NULL != client_cert) {
		sbuf = X509_NAME_oneline(X509_get_subject_name(client_cert), 0, 0);
		activity_write(BFHERE "Client certificate: %s", sbuf);
	}

	switch (ssl_state) {
		case SS_OK:
			BFTRACE("ssl_state = SS_OK");
			send_hello(CODE_HELLO);
			return NULL;
		case SS_SECURE:
			BFTRACE("ssl_state = SS_SECURE");
			send_hello(CODE_SECURE_HELLO);
			break;
		case SS_CERTIFIED:
			BFTRACE("ssl_state = SS_CERTIFIED");
			send_hello(CODE_CERTIFIED_HELLO);
			break;
		default:
			return NULL;
	}

	return ssl;
}


void report_SSL(void) {
	char *line;
	char *protocolVersion; 
	X509 *cert;

	if (!ssl) {
		send_msg("SSL", "SSLPlainText", NULL);
		return;
	}
	
	protocolVersion = SSL_get_version(ssl);
	activity_write(BFHERE "Protocol version: %s", protocolVersion);
	
    if(0 == strcmp(protocolVersion, "TLSv1") ){
       send_msg("SSL", "SSLStartTLS", "s", SSL_get_cipher_name(ssl));   
    }else if(0 == strcmp(protocolVersion, "TLSv1.1")){
       send_msg("SSL", "SSLStartTLSV11", "s", SSL_get_cipher_name(ssl));   
    }else if(0 == strcmp(protocolVersion, "TLSv1.2")){
       send_msg("SSL", "SSLStartTLSV12", "s", SSL_get_cipher_name(ssl));   
    }
    
    
	
	if (ssl_state != SS_CERTIFIED) {
		send_msg("SSL", "SSLUntrustedCert", NULL);
	}

	/* Get certificates (if available) */
	cert = SSL_get_peer_certificate(ssl);
	if (NULL != cert) {
		line = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0);
		if (NULL != line) {
			send_msg("SSL", "SSLPeerCertIssuer", "s", line);
		}
		line = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0);
		if (NULL != line) {
			send_msg("SSL", "SSLPeerCertSubject", "s", line);
		}
		X509_free(cert);
	}
}

#define io_read(buf, len) (ssl ? SSL_read(ssl, buf, len) : platform_protocol_read(buf, len))
#define io_write(buf, len) (ssl ? SSL_write(ssl, buf, len) : platform_protocol_write(buf, len))

#else /* !HAVE_SSL */

/*================================================================
 * Compiled without OpenSSL
 *----------------------------------------------------------------*/

void *start_SSL(void) {
	send_hello(CODE_HELLO);
	ssl_state = SS_OK;
	return NULL;
}

void stop_SSL(void) {
	send_goodbye();
}

void report_SSL(void) {
	/* Nothing to report, for now.  Might want to report that SSL is not
	 * being used.
	 */
}

#define io_read(buf, len) platform_protocol_read(buf, len)
#define io_write(buf, len) platform_protocol_write(buf, len)

#endif /* HAVE_SSL */

#endif /* HAVE_SYSTEM_SSL */



#define PUTS_BUF_SIZE (BFBUFSIZ << 2)
static char *puts_buf = NULL;


int io_puts(const char *s) {
	int len = (int)strlen(s);

	if (NULL != activity_file) {
		activity_write(BFHERE "IO io_puts: s =\"%s\"\n", s);
	}
	
	/* Writing UTF-8 to a socket ... */
	if (platform_protocol_has_socket() ||
			BFCONF_EXISTS(disable_transcode))
	{
#ifdef ZOS
		/* On z/OS, our UTF-8 data is transcoded to EBCDIC, and we
		 * need to trancode it back to real UTF-8 before we send it.
		 */
		int ret;
		char *p = malloc(len+1);
		memcpy(p, s, len+1);
		if (NULL != activity_file) {
			activity_write(BFHERE "IO io_puts: p =\"%s\"\n", p);
		}
		
		convert_etoa_l(p, len);
		
		/*
		if (NULL != activity_file) {
			activity_write(BFHERE "IO io_puts: convert_etoa_l: =\"\n%s\"", p);
		}
		*/
		ret = io_write(p, len);
		free(p);
		return ret;
#else /* !ZOS */
		/* On other platforms, UTF-8 is just UTF-8. */
		return io_write(s, len);
#endif /* ZOS */
	}

	/* Writing to the OS, so use its console encoding ... */
	if (platform_utf8_is_console()) {
		return io_write(s, len);
	}
	if (NULL == puts_buf) {
		puts_buf = malloc(PUTS_BUF_SIZE);
	}
	len = (int)convert_utf8_to_console(s, len, puts_buf, PUTS_BUF_SIZE);
	return io_write(puts_buf, len);
}


/* A very basic telnet protocol implementation.  It is rather common for
 * users to mistake telnet for a raw TCP/IP connection, but it actually
 * has binary escape sequences that prevent this from being true.  Since
 * we often recommend that people use telnet to test out the connection
 * to the agent, we need to tolerate any special codes their telnet client
 * sends us.  This adds some extra processing overhead, so to boost
 * performance, you might want to specify disable_telnet_support in
 * bfagent.conf to shut it off.  The drawback is that the agent will
 * hang up on any telnet client that sends any special codes to the agent,
 * as may commonly occur for option negotiation or non-ASCII characters.
 */
#define TELNET_SE   (240)
#define TELNET_NOP  (241)
#define TELNET_MARK (242)
#define TELNET_BRK  (243)
#define TELNET_IP   (244)
#define TELNET_AO   (245)
#define TELNET_AYT  (246)
#define TELNET_EC   (247)
#define TELNET_EL   (248)
#define TELNET_GA   (249)
#define TELNET_SB   (250)
#define TELNET_WILL (251)
#define TELNET_WONT (252)
#define TELNET_DO   (253)
#define TELNET_DONT (254)
#define TELNET_IAC  (255)

#define TNO_BINARY      (0)
#define TNO_SUPPRESS_GA (3)
#define TNO_TIMING_MARK (6)

#define GETS_UTF8_SIZE (BFBUFSIZ << 1)
static char gets_utf8[GETS_UTF8_SIZE];
static int gets_utf8_len = 0;

typedef enum {
	TNS_RESET,
	TNS_IAC,
	TNS_WILL,
	TNS_WONT,
	TNS_DO,
	TNS_DONT
} telnet_state_t;

static telnet_state_t telnet_state = TNS_RESET;
static const uint8_t TELNET_RESPONSE_AYT[] = {
	/* "\r\n[IBM Rational Build Forge Agent]\r\n" in ASCII/UTF-8. */
	0x0D, 0x0A, 0x5B,
	0x49, 0x42, 0x4D, 0x20,
	0x52, 0x61, 0x74, 0x69, 0x6F, 0x6E, 0x61, 0x6C, 0x20,
	0x42, 0x75, 0x69, 0x6C, 0x64, 0x20,
	0x46, 0x6F, 0x72, 0x67, 0x65, 0x20,
	0x41, 0x67, 0x65, 0x6E, 0x74,
	0x5D, 0x0D, 0x0A
};

/* We dump that string if we are explicitly asked for it by a
 * telnet request (hit control-] to escape to the telnet prompt
 * and type "send ayt"), or if we detect a serious violation of
 * the protocol.  The point is that either we've been asked to
 * identify ourselves or the person/software on the other end is
 * mistakenly talking to use as if we are some other program.
 */
int io_ayt(void) {
	return io_write(TELNET_RESPONSE_AYT, sizeof(TELNET_RESPONSE_AYT));
}

static void io_gets_telnet_option(uint8_t code) {

	uint8_t buf[3];
	buf[0] = TELNET_IAC;
	/* buf[1] = response */
	buf[2] = code;	

	switch (code) {
		/* Options we will claim to support, even though we don't actually
		 * do anything special for them.
		 */
		case TNO_BINARY:
		case TNO_SUPPRESS_GA:
		case TNO_TIMING_MARK:
		{
			switch (telnet_state) {
				case TNS_WILL: buf[1] = TELNET_DO;   break;
				case TNS_WONT: buf[1] = TELNET_DONT; break;
				case TNS_DO:   buf[1] = TELNET_WILL; break;
				case TNS_DONT: buf[1] = TELNET_WONT; break;
				default: /* NOT REACHED */ return;  
			}
			break;
		}

		/* We don't even understand what they're asking us to do,
		 * so refuse to cooperate.
		 */
		default:
		{
			switch (telnet_state) {
				case TNS_WILL: /* FALL THROUGH */
				case TNS_WONT: buf[1] = TELNET_DONT; break;
				case TNS_DO:   /* FALL THROUGH */
				case TNS_DONT: buf[1] = TELNET_WONT; break;
				default: /* NOT REACHED */ return;
			}
		}
	}
	io_write(buf, 3);
}

/* At some point, we discovered a 0xFF byte in the protocol stream.  That
 * is not a valid byte in UTF-8, so we will assume that a user is connected
 * to the agent using telnet and that it came from the telnet client.  This
 * is annoying, but we'll deal with it as best as we can by responding
 * appropriately when we see telnet control sequences.  If we encounter an
 * incomplete sequence, we'll save it for later.  Some sequences can be
 * quite long, so we give up if it looks like a sequence is too long to
 * bother trying to honor it.  In this case, we discard the IAC and
 * release the rest of the data to be processed as (alleged) UTF-8.
 */
static void io_gets_telnet(char *telnet_mark) {
	char *p = telnet_mark;
	char *pend = &gets_utf8[gets_utf8_len];

	/* Miniature FSM for telnet option negotiation */
	while (1) {
		/* Exhausted the buffer.  Truncate it at the telnet sequence. */
		if (p >= pend) {
			gets_utf8_len = (int)(telnet_mark - gets_utf8);
			return;
		}

		switch (telnet_state) {
			case TNS_RESET:
			{
				p = memchr(telnet_mark, TELNET_IAC, pend - telnet_mark);
				if (NULL == p)
					return;
				++p;
				telnet_state = TNS_IAC;
				continue;
			}

			case TNS_IAC:
			{
				switch ((uint8_t)(*p++)) {
					case TELNET_AYT:
						io_ayt();
						goto completed;
					case TELNET_EC:
						if (telnet_mark > gets_utf8)
							--telnet_mark;
						goto completed;
					case TELNET_EL:
						telnet_mark = memrchr(gets_utf8, 0x0A, gets_utf8_len);
						if (NULL == telnet_mark)
							telnet_mark = gets_utf8;
						goto completed;
					case TELNET_WILL:
						telnet_state = TNS_WILL;
						continue;
					case TELNET_WONT:
						telnet_state = TNS_WONT;
						continue;
					case TELNET_DO:
						telnet_state = TNS_DO;
						continue;
					case TELNET_DONT:
						telnet_state = TNS_DONT;
						continue;
					case TELNET_IAC:
						++telnet_mark;
						goto completed;
					default:
						goto completed;
				}
			}

			default: /* WILL WONT DO DONT */
				io_gets_telnet_option((uint8_t)*p++);
				goto completed;
		}

completed:
		gets_utf8_len -= (int)(p - telnet_mark);
		pend = &gets_utf8[gets_utf8_len];
		telnet_state = TNS_RESET;
		if (telnet_mark < &gets_utf8[gets_utf8_len]) {
			memmove(telnet_mark, p, pend - telnet_mark);
			p = telnet_mark = memchr(telnet_mark, TELNET_IAC, pend - telnet_mark);
			if (NULL != telnet_mark)
				continue;
		}
		return;
	}
}

/* Read UTF-8 data directly into the buffer ... */
static int io_gets_fill_utf8(void) { while (1) {
	int len;
	char *p = &gets_utf8[gets_utf8_len];
	char *q = p;

	len = io_read(p, GETS_UTF8_SIZE - gets_utf8_len);
	/*
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_utf8: io_read -> len=%d",
				len);
		activity_write(BFHERE
				"IO io_gets_fill_utf8: io_read -> content=\n%s",
				p);		
	}
	*/
	
	if (len <= 0) { return len; }
	
	p = &gets_utf8[gets_utf8_len];
	gets_utf8_len += len;

	/* If this is a person using a telnet, then we might see telnet
	 * control sequences.  This deals with them if there are any.
	 */
	if (!BFCONF_EXISTS(disable_telnet_support)) {
		if (telnet_state != TNS_RESET || NULL != (q = memchr(p, 0xFF, len))) {
			io_gets_telnet(q);
			len = (int)(&gets_utf8[gets_utf8_len] - p);
			if (len <= 0) { continue; }
		}
	}

#ifdef ZOS
	convert_atoe_l(p, len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_utf8: convert_atoe_l: \n%s\n",
				p);		
	}
#endif

	return gets_utf8_len;
}}


#define GETS_CONSOLE_SIZE (BFBUFSIZ)
/* Read from the OS, using its console encoding ... */
static char *gets_console = NULL;
static int gets_console_len = 0;

/* Reads from the console until there is a suitable boundary for
 * transcoding.  If the buffer fills completely and there is no
 * candidate boundary, then return the entire buffer.
 */
static int io_gets_fill_console(void) { while (1) {
	int len = io_read(&gets_console[gets_console_len],
			GETS_CONSOLE_SIZE - gets_console_len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_console: io_read -> len=%d",
				len);
	}

	/* The read returned an error or EOF.  If we have data buffered,
	 * then we flush it.  Otherwise, pass on the error/EOF result.
	 */
	if (len <= 0)
		return (gets_console_len > 0) ? gets_console_len : len;
	gets_console_len += len;

	/* See if we can find a good stopping point for transcoding. */
	len = (int)convert_console_find_boundary(gets_console, gets_console_len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_console: "
				"convert_console_find_boundary: %d -> %d",
				gets_console_len, len);
	}
	if (len > 0)
		return len;

	/* If the buffer is full, give up and transcode the whole thing. */
	if (gets_console_len >= GETS_CONSOLE_SIZE) {
		if (NULL != activity_file) {
			BFTRACE("IO io_gets_fill: console buffer overflow");
		}
		return GETS_CONSOLE_SIZE;
	}

	/* Not at a stopping point and still have room, so keep going... */
}}




static int io_gets_fill_transcode(void) {
	int len, converted;

	len = io_gets_fill_console();
	if (len <= 0) {
		if (gets_console_len <= 0) { return len; }
		len = gets_console_len;
	}

	/* Convert a block of console text to UTF-8. */
	converted = (int)convert_console_to_utf8(gets_console, len,
			&gets_utf8[gets_utf8_len], GETS_UTF8_SIZE - gets_utf8_len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_transcode: "
				"console [%d - %d] -> UTF8 [%d + %d]",
				gets_console_len, len,
				gets_utf8_len, converted);
	}

	gets_console_len -= len;
	memmove(gets_console, &gets_console[len], gets_console_len);
	return gets_utf8_len += converted;
}

static int (*io_gets_fill_fn)(void) = NULL;
#define io_gets_fill() ((*io_gets_fill_fn)())

#ifdef ZOS
#  define IS_NOT_CONTROL_CHAR(c) (convert_etoa_c(c) & ~0x1F)
# else /* !ZOS */
#  define IS_NOT_CONTROL_CHAR(c) ((c) & ~0x1F)
#endif /* ZOS/! */

char *io_gets(char *buf, int maxlen) {
	int i=0;
	char c;

	/* figure out if we are going to fill the buffer using a
	 * network socket or the console.
	 */
	if (NULL == io_gets_fill_fn) {
		if (platform_protocol_has_socket() ||
				BFCONF_EXISTS(disable_transcode)) {
			io_gets_fill_fn = &io_gets_fill_utf8;
		} else {
			io_gets_fill_fn = &io_gets_fill_transcode;
			gets_console = malloc(GETS_CONSOLE_SIZE);
		}
	}

	/* If we fill the buffer, then stop regardless of whether or not
	 * we find a newline.
	 */
	while (i < maxlen-1) {
		/* If we need more chars than are currently in the UTF-8 buffer,
		 * then we need to fill it.  If we can't, then return NULL.
		 */
		while (i >= gets_utf8_len) {
			if (io_gets_fill() <= 0) {
				if (NULL != activity_file) { BFTRACE("IO io_gets: NULL"); }
				return NULL;
			}
		}

		c = gets_utf8[i];
		/* Non-control character -- preserve and continue.  We also permit
		 * 0x0E and 0x0F, as they might show up in certain situations where
		 * the encoding is not UTF-8, but things would otherwise work.
		 */
		if (IS_NOT_CONTROL_CHAR(c) || c == 0x0E || c == 0x0F) {
			++i;
			continue;
		}

		/* Tolerate tab chars */
		if (c == '\t') {
			++i;
			continue;
		}

		/* Map CR to LF */
		if (c == '\r') { c = gets_utf8[i] = '\n'; }

		/* If it's \n, we have a full line. */
		if (c == '\n') {
			++i;
			break;
		}

		/* Backspace?  You must be using the broken Windows telnet...  We
		 * only honor it if there is something to delete; otherwise, we
		 * just discard it.
		 */
		if (c == '\b') {
			if (i > 0) {
				--i;
				gets_utf8_len -= 2;
				memmove(&gets_utf8[i], &gets_utf8[i+2], gets_utf8_len - i);
			} else {
				--gets_utf8_len;
				memmove(&gets_utf8[i], &gets_utf8[i+1], gets_utf8_len - i);
			}
			continue;
		}

		/* Any other control char -- immediately hang up on the client */
		activity_write(BFHERE "IO io_gets: invalid char 0x%02X", (unsigned int)(unsigned char)gets_utf8[i]);
		io_ayt();
		return NULL;
	}

	/* Return the data to the caller and remove it from the buffer. */
	memcpy(buf, gets_utf8, i);
	buf[i] = 0;
	gets_utf8_len -= i;
	memmove(gets_utf8, &gets_utf8[i], gets_utf8_len);

	/* Don't trust the client to speak UTF-8 properly... Some things, like
	 * overlong encoding, can be dangerous if they are allowed.
	 */
	convert_utf8_clean(buf, i, 1);

	
	if (NULL != activity_file) {
		activity_write(BFHERE "IO io_gets: [%d]=\"%s\"", i, buf);
	}
	
	return buf;
}


